/*
   ┏━━━┓╱╱╱╱╱┏┓╱╱┏┓╱╱╱╱╱┏┓╱╱   
   ┃┏━┓┃╱╱╱╱╱┃┃╱╱┃┃╱╱╱╱╱┃┃╱╱╱  
   ┃┗━┛┣┓┏┳━━┫┃┏┓┃┃╱╱┏━━┫┗━┳━━┓
   ┃┏━━┫┃┃┃┏┓┃┗┛┃┃┃╱┏┫┏┓┃┏┓┃━━┫
   ┃┃╱╱┃┗┛┃┃┃┃┏┓┃┃┗━┛┃┏┓┃┗┛┣━━┃
   ┗┛╱╱┗━━┻┛┗┻┛┗┛┗━━━┻┛┗┻━━┻━━┛
    ━━━━━━━━━━━━━━━━━━━━━━━━━━ 

	Copyright (c) 2022 Punk Labs LLC

	This file is part of OneTrick SIMIAN

	OneTrick SIMIAN is free software: you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation, either version 3 of the License, or (at your option)
	any later version.

	OneTrick SIMIAN is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	OneTrick SIMIAN.  If not, see <http://www.gnu.org/licenses/>.
*/

import "ui" for UI, Element, Point, Size, Rectangle, RGBA
import "widgets" for UIElement, UIWindowResizer, UIKnob, UIColorPanel, UITabContainer, UITab, UISeparator, UIPeakIndicator, UIPresetSelector, UICreditsPanel, UIButton, UIConnector, UIParameter, UIValue, UILabel, UIIcon

//Allowed @ScriptProperty types:
//bool, RGBA, ubyte, byte, ushort, short, uint, int, float, double, L16

/* Wren Notes:
	for(item in list) { }
	for(i in 0...100) { } //0-99
	for(i in 0..100) { } //0-100
*/


class Plugin {
	static liveEdit() { true }
	
	/*
		Reduce memory leaks from round-trips to native code
	*/
	static get(id) {
	/* //Maps don't work?
		if (__ElementCache.containsKey(id)) {
			return __ElementCache[id]
		}
		var result = get(id)
		__ElementCache[id] = result
		return result
	*/
		var index = __ElementCacheIDs.indexOf(id)
		if (index >= 0) {
			return __ElementCacheValues[index]
		}
		var result = $(id)
		__ElementCacheIDs.add(id)
		__ElementCacheValues.add(result)
		return result
		
	}

	// Color Correction
	//static correct(c) { c }
	//static correct(c) { RGBA.new(c.r, c.g, c.b, c.a) }
	static correct(c) { lightenColor(c, 0.1) }

	static initStyle() {
		//if (!__ElementCache) {
		//	__ElementCache = {} //Maps don't work?
		//}
		if (!__ElementCacheIDs) {
			__ElementCacheIDs = []
			__ElementCacheValues = []
		}
		__GlobalKnobNames = ["Gain", "Drive", "Sensitivity", "Saturation", "ReverbDecay", "ReverbGate"]
		__ModuleNames = ["module1", "module2", "module3", "module4", "module5", "module6", "module7", "module8", "module9", "module10"]
		__ModuleKnobNames = ["Gain", "Pan", "Noise", "Tone", "Bend", "Decay", "Punch", "ToneNoise", "Click", "GainDynamics", "BendDynamics", "LowpassQ", "Saturation", "FilterEnvBend", "SendReverb", "FilterEnvBendDynamics", "OscWaveform"]
		__ModuleButtonNames = ["Icon", "Color"]

		__gainStartClip = 0.708
		
		__colorClear = correct(RGBA.grey(255,0))
		__colorWhite = correct(RGBA.grey(255))
		__colorBlack = correct(RGBA.grey(0))
    	__colorGrey1 = correct(RGBA.grey(16))
    	__colorGrey2 = correct(RGBA.grey(32))
    	__colorGrey3 = correct(RGBA.grey(48))
    	__colorGrey4 = correct(RGBA.grey (64))
    	__colorGrey5 = correct(RGBA.grey(96))
    	__colorGrey6 = correct(RGBA.grey(128))
    	__colorGrey7 = correct(RGBA.grey(192))
    	__colorGrey8 = correct(RGBA.grey(224))
    	__colorGrey9 = correct(RGBA.grey(235))
    	__colorBrown = correct(RGBA.new(48,32,24,255))
    	
		__colorOrange = correct(RGBA.new(255, 155, 055, 255))
    	__colorBlue = correct(RGBA.new(55, 200, 255, 255))
		__colorPink = correct(RGBA.new(247, 61, 200, 255))
		__colorAqua = correct(RGBA.new(0, 255, 255, 255))
		__colorPurple = correct(RGBA.new(154, 118, 252, 255))
	   	__colorRed = correct(RGBA.new(252, 100, 100, 255))
		__colorYellow = correct(RGBA.new(252, 249, 100, 255))
		__colorGreen = correct(RGBA.new(100, 255, 100, 255))
		__colorClip = correct(RGBA.new(255, 64, 32, 255))
		__knobSize = 85
		
		__ModuleColors = [__colorRed, __colorOrange, __colorBlue, __colorAqua, __colorGreen, __colorPurple, __colorPink, __colorYellow]
	}
	static initLayout() {
      	//UI.defaultWidth
      	//UI.defaultHeight
      	//UI.root
      	//UI.getElementById("")
        __W = UI.width
        __DW = UI.defaultWidth
        __H = UI.height
        __DH = UI.defaultHeight
        __S = __W / __DW
	}
	
	static getModuleKnob(module, knobName) { get(module+"Knob"+knobName) }
	static getModuleButton(module, buttonName) { get(module+"Button"+buttonName) }
	static getModuleParam(module, paramName) { get(module+"Param"+paramName) }
	static getModuleValue(module, paramName) { get(module+"Value"+paramName) }
	static getModuleIndicator(module) { get(module+"Indicator") }
	static getModuleTab(module) { get(module+"Tab") }
	static getModulePanel(module) { get(module+"Panel") }
	static getModuleSeparator(module, name) { get(module+"Separator"+name) }
	static getModuleConnector(module, name) { get(module+"Connector"+name) }
	static getModuleLabel(module, name) { get(module+"Label"+name) }
	
	static flowPadding(r, padding) { Rectangle.new(r.origin.x+padding, r.origin.y+padding, r.size.width-padding*2, r.size.height-padding*2) }
	static flowPaddingH(r, padding) { Rectangle.new(r.origin.x+padding, r.origin.y, r.size.width-padding*2, r.size.height) }
	static flowPaddingV(r, padding) { Rectangle.new(r.origin.x, r.origin.y+padding, r.size.width, r.size.height-padding*2) }
	static flowPaddingTop(r, padding) { Rectangle.new(r.origin.x, r.origin.y+padding, r.size.width, r.size.height-padding) }
	static flowPaddingBottom(r, padding) { Rectangle.new(r.origin.x, r.origin.y, r.size.width, r.size.height-padding) }
	static flowPaddingLeft(r, padding) { Rectangle.new(r.origin.x+padding, r.origin.y, r.size.width-padding, r.size.height) }
	static flowPaddingRight(r, padding) { Rectangle.new(r.origin.x, r.origin.y, r.size.width-padding, r.size.height) }
	static flowOffset(r, x, y) { Rectangle.new(r.origin.x+x, r.origin.y+y, r.size.width, r.size.height) }
	
	static flowSplitH(r, i, count, spacing) {
		var chunkSize = (r.size.width - (spacing*(count-1)))/count
		var stride = chunkSize + spacing
		return Rectangle.new(r.origin.x+(stride*i), r.origin.y, chunkSize, r.size.height)
	}
	static flowSplitH(r, i, count) { flowSplitH(r, i, count, 0) }
	
	static flowSplitV(r, i, count, spacing) {
		var chunkSize = (r.size.height - (spacing*(count-1)))/count
		var stride = chunkSize + spacing
		return Rectangle.new(r.origin.x, r.origin.y+(stride*i), r.size.width, chunkSize)
	}
	static flowSplitV(r, i, count) { flowSplitV(r, i, count, 0) }
	
	static flowSetGrid(r, countX, countY) {
	__fsg_r = r
	__fsg_cx = countX
	__fsg_cy = countY
	}
	static flowUseGrid(iX, iY) { flowSplitV(flowSplitH(__fsg_r, iX, __fsg_cx), iY, __fsg_cy) }
	
	static flowGrid(r, iX, iY, countX, countY) { flowSplitV(flowSplitH(r, iX, countX), iY, countY) }
	static flowCenter(r, w, h) { Rectangle.new(r.origin.x+(r.size.width-w)/2, r.origin.y+(r.size.height-h)/2, w, h) }
	static flowCenter(r, s) { Rectangle.new(r.origin.x+(r.size.width-s)/2, r.origin.y+(r.size.height-s)/2, s, s) }
	static flowCenterW(r, w) { Rectangle.new(r.origin.x+(r.size.width-w)/2, r.origin.y, w, r.size.height) }
	static flowCenterH(r, h) { Rectangle.new(r.origin.x, r.origin.y+(r.size.height-h)/2, r.size.width, h) }
	static flowTop(r, h) { Rectangle.new(r.origin.x, r.origin.y, r.size.width, h) }
	static flowBottom(r, h) { Rectangle.new(r.origin.x, r.origin.y+r.size.height-h, r.size.width, h) }
	static flowLeft(r, w) { Rectangle.new(r.origin.x, r.origin.y, w, r.size.height) }
	static flowRight(r, w) { Rectangle.new(r.origin.x+r.size.width-w, r.origin.y, w, r.size.height) }
	static flowWidth(r, w) { Rectangle.new(r.origin.x, r.origin.y, w, r.size.height) }
	static flowHeight(r, h) { Rectangle.new(r.origin.x, r.origin.y, r.size.width, h) }
	
	static darkenColor(color, t) { RGBA.new(color.r*t,color.g*t,color.b*t,color.a) }
	static lightenColor(color, t) { RGBA.new(color.r+(255-color.r)*t,color.g+(255-color.g)*t,color.b+(255-color.b)*t,color.a) }
	static blendColor(a, b, t) { RGBA.new(a.r+(b.r-a.r)*t, a.g+(b.g-a.g)*t, a.b+(b.b-a.b)*t, a.a+(b.a-a.a)*t) }

	
	static styleKnobBasic(knob) {
		//knob.colorBacking = correct(RGBA.new(0,0,0,0))
		//knob.margin = 0.05
		knob.indicatorStyle = 0
		knob.indicatorOrigin = 0
		if (knob.indicatorStyle == 0) {
			knob.indicatorWidth = 0.1
		} else {
			knob.indicatorWidth = 0.12
			knob.indicatorLedCount = 25
		}
		knob.indicatorGlow = 0.075
		knob.indicatorGlowRadius=2

		knob.knobRadius = 0.75
		knob.hoverScale = 1.05
		knob.colorKnob = __colorWhite
		//knob.colorKnobFade = correct(RGBA.new(255,255,255,16))
		knob.colorKnobFade = correct(RGBA.new(10,22,55,32))
		knob.colorKnobFadeHover = lightenColor(knob.colorKnobFade, 0.5)//correct(RGBA.new(100,200,255,92))
		knob.knobFadeR1=0.33
		knob.knobFadeR2=0.93
		knob.knobFadeR3=0.98
		knob.colorIndicator = __colorGrey4
		knob.colorHover = __colorWhite
		knob.colorDragged = __colorYellow
		knob.labelColor = __colorWhite
		knob.labelBackgroundColor = __colorClear //__colorWhite.withAlpha(32)
		knob.labelValueColor = __colorGrey6

		knob.jogStyle = 1
		if (knob.jogStyle == 0) {
			knob.jogOffset = 0.75
			knob.jogCircleRadius = 0.1
		} else {
			knob.jogStyle = 1
			knob.jogOffset = 0.9
			knob.jogWidth = 0.125
			knob.jogLength = 0.4
		}
	}
	static styleKnobDark(knob) {
		if(true) {
			knob.colorBacking = correct(RGBA.new(32,40,48,255))
			knob.indicatorStyle = 1
			knob.colorIndicator = __colorGrey5
			knob.colorIndicatorFill = __colorYellow
			knob.labelColor = __colorWhite //__colorBlack
			knob.labelValueColor = __colorGrey2
			knob.colorJog = __colorGrey2
			//knob.indicatorWidth = 0.1
			knob.margin = 0.1
		} else {
			knob.colorBacking = correct(RGBA.new(0,16,32,64))
			knob.indicatorStyle = 0
			knob.colorKnob = __colorGrey2
			knob.colorKnobFade = correct(RGBA.new(0,0,0,32))
			knob.colorKnobFadeHover = darkenColor(knob.colorKnobFade, 0.5)//correct(RGBA.new(100,200,255,92))
			knob.colorIndicator = __colorGrey4
			knob.colorIndicatorFill = correct(RGBA.new(255,0,255,255))
			knob.colorHover = __colorWhite.withAlpha(64)
			knob.colorDragged = correct(RGBA.new(255,0,255,255))
			knob.labelColor = __colorBlack
			knob.indicatorGlow = 0
		}
	}
	
	static styleModule(moduleName, color) {
	    color = __ModuleColors[getModuleParam(moduleName, "Color").intValue % 8]
	
		//getModulePanel(moduleName).color =  darkenColor(color, 0.1)
		getModulePanel(moduleName).color = __colorGrey2 // blendColor(darkenColor(color, 0.7), __colorGrey1, 0.8)
		
		var tab = getModuleTab(moduleName)
		if (tab) {
			var dColor = darkenColor(color,0.85)
			tab.colorFillInactive = __colorClear
			tab.colorFillActive = __colorWhite
			tab.colorFillHover = __colorWhite
			tab.colorOutlineInactive = __colorGrey6
			tab.colorOutlineActive = dColor
			tab.colorOutlineHover = dColor
			tab.colorLabelActive = dColor
			tab.colorLabelHover = dColor
			tab.styleOutlineThickness = 0.05
			tab.labelOffsetY = -7*__S
			//tab.labelOffsetY = 0
			//tab.labelRelativeHeight = 0.7
			tab.labelRelativeHeight = 0
		}
		
		var indicator=getModuleIndicator(moduleName)
		if (indicator) {
			indicator.oneShot = true
			indicator.peakHoldMS = 16.67*3
			indicator.peakFallMS = 225
			indicator.color = __colorGrey3 //darkenColor(color, 0.5)
			indicator.colorActive = darkenColor(color, 0.7)
			indicator.colorPeaking = lightenColor(color, 0.5)
			//if ( ["module3", "module4", "module5", "module6", "module10"].indexOf(moduleName)>=0 ) {
			if (!getModuleValue(moduleName, "Enabled").boolValue) {
				indicator.color = __colorGrey7
				indicator.colorActive = __colorGrey7
				indicator.colorPeaking = __colorGrey7
			}
			indicator.scale = 0.8
			indicator.scalePeaking = 0.97
		}
		
		for (knobName in __ModuleKnobNames) {
			var knob = getModuleKnob(moduleName, knobName)
			styleKnobBasic(knob)
			knob.colorIndicatorFill = color
			knob.labelColor = color
			knob.colorJog = darkenColor(color, 0.33)
		}
		
		for (knobName in ["Pan", "ToneNoise", "Bend", "FilterEnvBend"]) {
			var knob = getModuleKnob(moduleName, knobName)
			knob.indicatorOrigin = 1
			knob.indicatorDetent = true //(knobName=="Bend")
		}
		
		for (knobName in ["GainDynamics", "BendDynamics", "FilterEnvBendDynamics"]) {
			var knob = getModuleKnob(moduleName, knobName)
			knob.indicatorStyle = 1
			knob.colorIndicatorFill = lightenColor(color, 0.75)
		}
		
		/*
		for (knobName in ["SendReverb"]) {
			var knob = getModuleKnob(moduleName, knobName)
			knob.colorIndicatorFill = __colorPurple
		}
		*/
		
		for (i in 1..4) {
			getModuleSeparator(moduleName, i.toString).color = __colorGrey9
		}
		for (i in 1..5) {
			var label = getModuleLabel(moduleName, i.toString)
			label.color = __colorGrey9
			//label.colorShadow = correct(RGBA.new(0,0,0,96))
			label.shadowOffset = 0.1
		}
		for (i in 1..3) {
			var connector = getModuleConnector(moduleName, i.toString)
			connector.lineWidth = 3*__S
			connector.style = 0
			connector.color = __colorGrey6
			connector.colorEditing = __colorYellow
		}
		/*
		for (i in 4..5) {
			var connector = getModuleConnector(moduleName, i.toString)
			connector.style = 3
		}
		*/
		{
			var knob = getModuleKnob(moduleName, "Gain")
			knob.indicatorClipStart = __gainStartClip
			knob.colorIndicatorClip = __colorClip
		}
		
		for(buttonName in __ModuleButtonNames) {
			var button = getModuleButton(moduleName, buttonName)
			button.color = __colorGrey5
			button.colorBackgroundHover = __colorClear
			button.colorHover = __colorPink
			button.colorPressed = __colorYellow
		}

			

	}
	
	static reflowKnob(moduleName, knobName, x, y, w, h) {
		var knob  = getModuleKnob(moduleName, knobName)
		if (knob) {
			knob.position = Rectangle.new(x, y, w, h)
		}
	}
	
	static reflowModule(moduleName, x, y, w, h) {
		var panel = getModulePanel(moduleName)
		if (panel) {
			panel.position = Rectangle.new(x,y,w,h)
		}
		var index = __ModuleNames.indexOf(moduleName)
		
		var indicator=getModuleIndicator(moduleName)
		if (indicator) {
			//var tab = getModuleTab(moduleName)
			//if (tab) { indicator.position = flowCenter(tab.position, 50*__S) }
			var spacing = (w/__ModuleNames.count)
			var indicatorX = (w/__ModuleNames.count)*index
			var indicatorSize = 50*__S
			var offset = -indicatorSize/2
			var offsetX = spacing / 2
			indicator.position = Rectangle.new(indicatorX + offsetX + offset, y - (32*__S) + offset, indicatorSize, indicatorSize)
		}
		
		{
			var fullR = flowPaddingH(Rectangle.new(x,y,w,h), 20*__S)
			var r = flowPaddingV(fullR, 40*__S)
			
			var gx = 11
			var gy = 2
			var gs = __knobSize*__S
			flowSetGrid(r, gx, gy)
			
			//get("module1ButtonIcon").position = Rectangle.new(0,0,10,10)
			getModuleButton(moduleName, "Icon").position = Rectangle.new(fullR.origin.x+10*__S, fullR.origin.y, 30*__S, 30*__S)
			getModuleButton(moduleName, "Color").position = Rectangle.new(fullR.origin.x-20*__S, fullR.origin.y, 30*__S, 30*__S)
			
			var ix = 0
			getModuleKnob(moduleName, "Tone").position = flowCenter(flowUseGrid(ix, 0), gs)
			getModuleKnob(moduleName, "Bend").position = flowCenter(flowUseGrid(ix+1, 0), gs)
			getModuleKnob(moduleName, "OscWaveform").position = flowCenter(flowUseGrid(ix, 1), gs)
			getModuleKnob(moduleName, "BendDynamics").position = flowCenter(flowUseGrid(ix+1, 1), gs)
			
			ix = ix + 2.5
			getModuleKnob(moduleName, "Noise").position = flowCenter(flowUseGrid(ix, 0), gs)
			getModuleKnob(moduleName, "FilterEnvBend").position = flowCenter(flowUseGrid(ix+1, 0), gs)
			getModuleKnob(moduleName, "LowpassQ").position = flowCenter(flowUseGrid(ix, 1), gs)
			getModuleKnob(moduleName, "FilterEnvBendDynamics").position = flowCenter(flowUseGrid(ix+1, 1), gs)
			
			ix = ix + 2.5
			getModuleKnob(moduleName, "Decay").position = flowCenter(flowUseGrid(ix, 0), gs)
			getModuleKnob(moduleName, "Punch").position = flowCenter(flowUseGrid(ix, 1), gs)

			ix = ix + 1.5
			getModuleKnob(moduleName, "Pan").position = flowCenter(flowUseGrid(ix, 0), gs)
			getModuleKnob(moduleName, "SendReverb").position = flowCenter(flowUseGrid(ix+1, 0), gs)
			getModuleKnob(moduleName, "ToneNoise").position = flowCenter(flowUseGrid(ix, 1), gs)
			getModuleKnob(moduleName, "Click").position = flowCenter(flowUseGrid(ix+1, 1), gs)

			ix = ix + 2.5
			getModuleKnob(moduleName, "Gain").position = flowCenter(flowUseGrid(ix+1, 0), gs)
			getModuleKnob(moduleName, "Saturation").position = flowCenter(flowUseGrid(ix, 0.5), gs)
			getModuleKnob(moduleName, "GainDynamics").position = flowCenter(flowUseGrid(ix+1, 1), gs)

			var sw = 2*__S
			var sh = 300*__S
			getModuleSeparator(moduleName, "1").position = flowCenter(flowSplitH(r, 1.75, gx), sw, sh)
			getModuleSeparator(moduleName, "2").position = flowCenter(flowSplitH(r, 4.25, gx), sw, sh)
			getModuleSeparator(moduleName, "3").position = flowCenter(flowSplitH(r, 5.75, gx), sw, sh)
			getModuleSeparator(moduleName, "4").position = flowCenter(flowSplitH(r, 8.25, gx), sw, sh)
			var cw = 10*__S
			var ch = 130*__S
			getModuleConnector(moduleName, "1").position = flowCenter(flowSplitH(r, 1.6, gx), cw, ch)
			getModuleConnector(moduleName, "2").position = flowCenter(flowSplitH(r, 4.1, gx), cw, ch)
			getModuleConnector(moduleName, "3").position = flowCenter(flowSplitH(r, 10.6, gx), cw, ch)
			// getModuleConnector(moduleName, "4").position = flowTop(flowUseGrid(0.5, 0.05), 8*__S)
			// getModuleConnector(moduleName, "5").position = flowTop(flowUseGrid(3.0, 0.05), 8*__S)

			
			var labelR = flowTop(flowPaddingV(fullR, 20*__S), 25*__S)
			getModuleLabel(moduleName, "1").position = flowCenterW(flowSplitH(labelR, 0.5, gx), 100*__S)
			getModuleLabel(moduleName, "2").position = flowCenterW(flowSplitH(labelR, 3, gx), 100*__S)
			getModuleLabel(moduleName, "3").position = flowCenterW(flowSplitH(labelR, 5, gx), 100*__S)
			getModuleLabel(moduleName, "4").position = flowCenterW(flowSplitH(labelR, 7, gx), 100*__S)
			getModuleLabel(moduleName, "5").position = flowCenterW(flowSplitH(labelR, 9.5, gx), 100*__S)
		

		}
	}

	static parameterChanged() {
		/* //Allocation Stress Test
		for (i in 0...1000) {
			for (moduleName in __ModuleNames) {
				for (knobName in __ModuleKnobNames) {
					//var knob = getModuleKnob(moduleName, knobName)
					var strTest = moduleName+"Knob"+knobName+"a"+"b"+"c"
				}
			}
		}
		*/
	    styleModules()
	    
	    System.gc()
	}
	
	static styleModules() {
	    styleModule("module1", __colorRed)
	    styleModule("module2", __colorOrange)
	    styleModule("module3", __colorOrange)
	    styleModule("module4", __colorOrange)
	    styleModule("module5", __colorBlue)
	    styleModule("module6", __colorAqua)
	    styleModule("module7", __colorGreen)
	    styleModule("module8", __colorPink)
	    styleModule("module9", __colorPurple)
	    styleModule("module10", __colorYellow)
	}
	
	static styleUI() {
        //System.print("plugin.createUI()")
		initStyle()
		initLayout()
		
		get("background").color = darkenColor(__colorAqua, 0.9) //__colorGrey7
		//get("background").color = darkenColor(__ModuleColors[getModuleParam("module1", "Color").intValue % 8], 0.9)
		
        
		for (logo in [get("logoLabel"), get("logoIcon"), get("simianIcon")]) {
			logo.color = __colorGrey2
	        logo.colorShadow = __colorClear
			logo.shadowOffset = 0.035
		}
		get("simianIcon").color = __colorGrey8

        var versionLabel = get("versionLabel")
        if (versionLabel) {
        	versionLabel.color = __colorGrey5
        	//versionLabel.colorBackground = __colorWhite
        }

		
		var tabContainer = get("tabContainer")
		tabContainer.colorFill = __colorGrey9
		
		for (moduleName in __ModuleNames) {
			getModulePanel(moduleName).color = __colorGrey5
			}

		styleModules()
	    
	    //UI.root.color = correct(RGBA.new(255,255,255,255))

		for (button in [get("creditsButton"), get("upgradeButton")]) {
			button.colorBackground = __colorClear
			button.colorBackgroundHover = __colorClear
			button.colorBackgroundPressed = __colorClear
			button.color = __colorWhite //__colorGrey5 //.withAlpha(128) //Alpha not working on fonts/icons yet
			button.colorHover = __colorYellow
			button.colorPressed = __colorYellow
			button.colorShadow = blendColor(__colorAqua, __colorBlack, 0.333)
			button.shadowOffset = 0.075
		}
		get("upgradeButton").color = __colorYellow
		get("upgradeButton").colorHover = __colorAqua


        //get("globalPanel").color = __colorClear
        get("globalPanel").color = darkenColor(__colorRed, 0.8)
        //get("globalPanel").color = darkenColor(__ModuleColors[getModuleParam("module1", "Color").intValue % 8], 0.8)

        
        for (knobName in __GlobalKnobNames) {
        	var knob = getModuleKnob("global", knobName)
			styleKnobBasic(knob)
			styleKnobDark(knob)
        }
        var knob = getModuleKnob("global", "Gain")
		knob.indicatorStyle = 0
		knob.indicatorClipStart = __gainStartClip
		knob.colorIndicatorClip = __colorClip

        knob = getModuleKnob("global", "Drive")
		knob.indicatorStyle = 0
		knob.indicatorClipStart = __gainStartClip
		knob.colorIndicatorClip = __colorClip

        knob = getModuleKnob("global", "Sensitivity")
		knob.colorIndicatorFill = __colorOrange
		knob.indicatorStyle = 1
		
        knob = getModuleKnob("global", "Saturation")
		knob.colorIndicatorFill = __colorPink
		knob.indicatorStyle = 0

        knob = getModuleKnob("global", "ReverbDecay")
		knob.colorIndicatorFill = __colorPurple
		knob.indicatorStyle = 0
		
        knob = getModuleKnob("global", "ReverbGate")
		knob.colorIndicatorFill = __colorPurple
		knob.indicatorStyle = 0

	    
	    var resizer = get("resizerHint")
	    resizer.color = __colorGrey9
	    resizer.colorHovered = __colorYellow
	    resizer.colorDragged = __colorYellow
	    resizer.colorCannotResize = __colorRed
	    
	}


    static createUI() {
		styleUI()
    }

    static reflow() {
		initLayout()

    	if (liveEdit()) {
	    	styleUI()
	    }

		//System.print("plugin.reflow()")
		
      	
        get("background").position = Rectangle.new(0,0,__W,__H)
        get("logoIcon").position = Rectangle.new(10,10,60,60).scaleByFactor(__S)
        get("logoLabel").position = Rectangle.new(70,15,160,50).scaleByFactor(__S)
        get("simianIcon").position = Rectangle.new(802,467,180,180).scaleByFactor(__S)
        var versionLabel = get("versionLabel")
        if (versionLabel) {
        	versionLabel.position = Rectangle.new(70,58,160,22).scaleByFactor(__S)
        }
		var buttonSize = 35*__S
		get("creditsButton").position = Rectangle.new(240*__S, 10*__S, buttonSize, buttonSize)
        get("presetSelector").position = Rectangle.new(600,25,380,30).scaleByFactor(__S)
        
        get("upgradeButton").position = Rectangle.new(0,515,112,140).scaleByFactor(__S)

        //get("GG").position = Rectangle.new(__DW-70, 6, 64, 64).scaleByFactor(__S)
        //UI.getElementById("GG").position = Rectangle.new(__DW-70, 6, 64, 64).scaleByFactor(__S)
        //UIElement.new("GG").position = Rectangle.new(__DW-70, 6, 64, 64).scaleByFactor(__S)
        
	    get("creditsPanel").textSize = 25 * __S

        
        var y = 80*__S
        var tabHeight = 64*__S
        var tabContainer = get("tabContainer")
		tabContainer.horizontalSkew = 0 //-20*__S
        tabContainer.position = Rectangle.new(0, y, __W, tabHeight)

		y = y + tabHeight
		var panelHeight = 330*__S
		var i = 0
		for (moduleName in __ModuleNames) {
			reflowModule(moduleName, 0, y, __W, panelHeight)
			i=i+1
		}
		y = y + panelHeight
		var gs = __knobSize*__S
		
		panelHeight = __H-y
		var fullR = Rectangle.new(0, y, __W, panelHeight)
		var r = fullR
		get("globalPanel").position = r
		//r = Rectangle.new(r.origin.x, r.origin.y-15*__S, r.size.width, r.size.height)
		var gx = 11
		r = flowPaddingH(r, 20*__S)
		flowSetGrid(r, gx, 1)
		var offset = 1 //1.75
        getModuleKnob("global", "Sensitivity").position = flowCenter(flowUseGrid(0.0+offset, 0), gs)
        getModuleKnob("global", "ReverbDecay").position = flowCenter(flowUseGrid(1.5+offset, 0), gs)
        getModuleKnob("global", "ReverbGate").position = flowCenter(flowUseGrid(2.5+offset, 0), gs)
        getModuleKnob("global", "Drive").position = flowCenter(flowUseGrid(4.0+offset, 0), gs)
        getModuleKnob("global", "Saturation").position = flowCenter(flowUseGrid(5.0+offset, 0), gs)
        getModuleKnob("global", "Gain").position = flowCenter(flowUseGrid(6.0+offset, 0), gs)
		var sw = 2*__S
		var sh = 140*__S
		getModuleSeparator("global", "1").position = flowCenter(flowSplitH(r, 0.75+offset, gx), sw, sh)
		getModuleSeparator("global", "2").position = flowCenter(flowSplitH(r, 3.25+offset, gx), sw, sh)
		var labelR = flowTop(flowPaddingV(r, 5*__S), 25*__S)
		getModuleLabel("global", "1").position = flowCenterW(flowSplitH(labelR, 0.0+offset, gx), 100*__S)
		getModuleLabel("global", "2").position = flowCenterW(flowSplitH(labelR, 2.0+offset, gx), 100*__S)
		getModuleLabel("global", "3").position = flowCenterW(flowSplitH(labelR, 5.0+offset, gx), 100*__S)

        get("resizerHint").position = Rectangle.new(__W-30, __H-30, 30, 30)
        
        System.gc()
    }
}
