/*
   ┏━━━┓╱╱╱╱╱┏┓╱╱┏┓╱╱╱╱╱┏┓╱╱   
   ┃┏━┓┃╱╱╱╱╱┃┃╱╱┃┃╱╱╱╱╱┃┃╱╱╱  
   ┃┗━┛┣┓┏┳━━┫┃┏┓┃┃╱╱┏━━┫┗━┳━━┓
   ┃┏━━┫┃┃┃┏┓┃┗┛┃┃┃╱┏┫┏┓┃┏┓┃━━┫
   ┃┃╱╱┃┗┛┃┃┃┃┏┓┃┃┗━┛┃┏┓┃┗┛┣━━┃
   ┗┛╱╱┗━━┻┛┗┻┛┗┛┗━━━┻┛┗┻━━┻━━┛
    ━━━━━━━━━━━━━━━━━━━━━━━━━━ 

	Copyright (c) 2022 Punk Labs LLC

	This file is part of OneTrick SIMIAN

	OneTrick SIMIAN is free software: you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation, either version 3 of the License, or (at your option)
	any later version.

	OneTrick SIMIAN is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	OneTrick SIMIAN.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
	This class is a quick-and-dirty port of the Wren script used during development.
	Doing layout here should be faster and less leaky.
*/

module guistyle;

import dplug.gui.element;
import dplug.flatwidgets;

import punklabs.ui.button;
import punklabs.ui.customgui;
import punklabs.ui.colorpanel;
import punklabs.ui.connector;
import punklabs.ui.creditspanel;
import punklabs.ui.helper;
import punklabs.ui.knob;
import punklabs.ui.label;
import punklabs.ui.icon;
import punklabs.ui.pager;
import punklabs.ui.panel;
import punklabs.ui.peakindicator;
import punklabs.ui.presetselector;
import punklabs.ui.separator;
import punklabs.ui.tabcontainer;
import punklabs.ui.tab;
import punklabs.ui.parameter;
import punklabs.ui.value;
import punklabs.string.helper;


class GUIStyle {
public:
nothrow:
@nogc:
	struct Coord
	{
		float x;
		float y;
	}
	struct Size
	{
		float width;
		float height;
	}
	struct Rectangle
	{
	public:
	nothrow:
	@nogc:
		float x;
		float y;
		float width;
		float height;
		@property Coord origin() { return Coord(x, y); }
		@property Size size() { return Size(width, height); }
		this(float x, float y, float width, float height)
		{
			assert(x>=0);
			assert(y>=0);
			this.x = x;
			this.y = y;
			this.width = width;
			this.height = height;
		}
		Rectangle scaleByFactor(float scale)
		{
			return Rectangle(x*scale, y*scale, width*scale, height*scale);
		}
		
		@property box2i toBox()
		{
			return box2i(cast(int)(x+0.5f),cast(int)(y+0.5f),cast(int)(x+width+0.5f),cast(int)(y+height+0.5f));
		}
    	alias toBox this;
	}
	
	UIElement rootElement = null;
	this(UIElement rootElement)
	{
		this.rootElement = rootElement;
		createUI();
	}
	
	float W;
	float DW;
	float H;
	float DH;
	float S;

	bool initLayout() {
		if(rootElement.position.width < 30 || rootElement.position.height < 30)
			return false;
        W = rootElement.position.width;
        DW = rootElement.context.getDefaultUIWidth();
        H = rootElement.position.height;
        DH = rootElement.context.getDefaultUIHeight();
        S = W / DW;
        return true;
	}

	
	
	
	UIElement get(string id) {
		return rootElement.getElementById(cast(char*)id);
	}
	UIElement get(string id1, string id2) {
		char[1000] combined;
		StringHelper.formatString(combined, "%s%s", cast(char*)id1, cast(char*)id2);
		return get(cast(string)combined);
	}
	UIElement get(string id1, string id2, string id3) {
		char[1000] combined;
		StringHelper.formatString(combined, "%s%s%s", cast(char*)id1, cast(char*)id2, cast(char*)id3);
		return get(cast(string)combined);
	}

	static RGBA correct(RGBA c) { return lightenColor(c, 0.1); }

	const string[] __GlobalKnobNames = ["Gain", "Drive", "Sensitivity", "Saturation", "ReverbDecay", "ReverbGate"];
	const string[] __ModuleNames = ["module1", "module2", "module3", "module4", "module5", "module6", "module7", "module8", "module9", "module10"];
	const string[] __ModuleKnobNames = ["Gain", "Pan", "Noise", "Tone", "Bend", "Decay", "Punch", "ToneNoise", "Click", "GainDynamics", "BendDynamics", "LowpassQ", "Saturation", "FilterEnvBend", "SendReverb", "FilterEnvBendDynamics", "OscWaveform"];
	const string[] __ModuleButtonNames = ["Icon", "Color"];
	const string[] __Numbers = ["0","1","2","3","4","5","6","7","8","9","10"]; // quick hack

	const float __gainStartClip = 0.708;
	
	static RGBA RGBA_new(ubyte rgb) { return RGBA(rgb, rgb, rgb, 255); }
	static RGBA RGBA_new(ubyte rgb, ubyte a) { return RGBA(rgb, rgb, rgb, a); }
	static RGBA RGBA_new(ubyte r, ubyte g, ubyte b) { return RGBA(r, g, b, 255); }
	static RGBA RGBA_new(ubyte r, ubyte g, ubyte b, ubyte a) { return RGBA(r, g, b, a); }
		
	const RGBA __colorClear = correct(RGBA_new(255,0));
	const RGBA __colorWhite = correct(RGBA_new(255));
	const RGBA __colorBlack = correct(RGBA_new(0));
    const RGBA __colorGrey1 = correct(RGBA_new(16));
    const RGBA __colorGrey2 = correct(RGBA_new(32));
    const RGBA __colorGrey3 = correct(RGBA_new(48));
    const RGBA __colorGrey4 = correct(RGBA_new(64));
    const RGBA __colorGrey5 = correct(RGBA_new(96));
    const RGBA __colorGrey6 = correct(RGBA_new(128));
    const RGBA __colorGrey7 = correct(RGBA_new(192));
    const RGBA __colorGrey8 = correct(RGBA_new(224));
    const RGBA __colorGrey9 = correct(RGBA_new(235));
    const RGBA __colorBrown = correct(RGBA_new(48,32,24,255));
    	
	const RGBA __colorOrange = correct(RGBA_new(255, 155, 55, 255));
    const RGBA __colorBlue = correct(RGBA_new(55, 200, 255, 255));
	const RGBA __colorPink = correct(RGBA_new(247, 61, 200, 255));
	const RGBA __colorAqua = correct(RGBA_new(0, 255, 255, 255));
	const RGBA __colorPurple = correct(RGBA_new(154, 118, 252, 255));
	const RGBA __colorRed = correct(RGBA_new(252, 100, 100, 255));
	const RGBA __colorYellow = correct(RGBA_new(252, 249, 100, 255));
	const RGBA __colorGreen = correct(RGBA_new(100, 255, 100, 255));
	const RGBA __colorClip = correct(RGBA_new(255, 64, 32, 255));
	const float __knobSize = 85;
		
	RGBA[] __ModuleColors = [__colorRed, __colorOrange, __colorBlue, __colorAqua, __colorGreen, __colorPurple, __colorPink, __colorYellow];
	
	
	
	UIKnob getModuleKnob(string m, string knobName) { return cast(UIKnob)get(m,"Knob",knobName); }
	UIButton getModuleButton(string m, string buttonName) { return cast(UIButton)get(m,"Button",buttonName); }
	UIParameter getModuleParam(string m, string paramName) { return cast(UIParameter)get(m,"Param",paramName); }
	UIValue getModuleValue(string m, string paramName) { return cast(UIValue)get(m,"Value",paramName); }
	UIPeakIndicator getModuleIndicator(string m) { return cast(UIPeakIndicator)get(m,"Indicator"); }
	UITab getModuleTab(string m) { return cast(UITab)get(m,"Tab"); }
	UIColorPanel getModulePanel(string m) { return cast(UIColorPanel)get(m,"Panel"); }
	UISeparator getModuleSeparator(string m, string name) { return cast(UISeparator)get(m,"Separator",name); }
	UIConnector getModuleConnector(string m, string name) { return cast(UIConnector)get(m,"Connector",name); }
	UILabel getModuleLabel(string m, string name) { return cast(UILabel)get(m,"Label",name); }
	
	Rectangle flowPadding(Rectangle r, float padding) { return Rectangle(r.origin.x+padding, r.origin.y+padding, r.size.width-padding*2, r.size.height-padding*2); }
	Rectangle flowPaddingH(Rectangle r, float padding) { return Rectangle(r.origin.x+padding, r.origin.y, r.size.width-padding*2, r.size.height); }
	Rectangle flowPaddingV(Rectangle r, float padding) { return Rectangle(r.origin.x, r.origin.y+padding, r.size.width, r.size.height-padding*2); }
	Rectangle flowPaddingTop(Rectangle r, float padding) { return Rectangle(r.origin.x, r.origin.y+padding, r.size.width, r.size.height-padding); }
	Rectangle flowPaddingBottom(Rectangle r, float padding) { return Rectangle(r.origin.x, r.origin.y, r.size.width, r.size.height-padding); }
	Rectangle flowPaddingLeft(Rectangle r, float padding) { return Rectangle(r.origin.x+padding, r.origin.y, r.size.width-padding, r.size.height); }
	Rectangle flowPaddingRight(Rectangle r, float padding) { return Rectangle(r.origin.x, r.origin.y, r.size.width-padding, r.size.height); }
	Rectangle flowOffset(Rectangle r, float x, float y) { return Rectangle(r.origin.x+x, r.origin.y+y, r.size.width, r.size.height); }
	
	Rectangle flowSplitH(Rectangle r, float i, float count, float spacing) {
		auto chunkSize = (r.size.width - (spacing*(count-1)))/count;
		auto stride = chunkSize + spacing;
		return Rectangle(r.origin.x+(stride*i), r.origin.y, chunkSize, r.size.height);
	}
	Rectangle flowSplitH(Rectangle r, float i, float count) { return flowSplitH(r, i, count, 0); }
	
	Rectangle flowSplitV(Rectangle r, float i, float count, float spacing) {
		auto chunkSize = (r.size.height - (spacing*(count-1)))/count;
		auto stride = chunkSize + spacing;
		return Rectangle(r.origin.x, r.origin.y+(stride*i), r.size.width, chunkSize);
	}
	Rectangle flowSplitV(Rectangle r, float i, float count) { return flowSplitV(r, i, count, 0); }

	Rectangle __fsg_r;
	float __fsg_cx, __fsg_cy;
	void flowSetGrid(Rectangle r, float countX, float countY) {
		__fsg_r = r;
		__fsg_cx = countX;
		__fsg_cy = countY;
	}
	Rectangle flowUseGrid(float iX, float iY) { return flowSplitV(flowSplitH(__fsg_r, iX, __fsg_cx), iY, __fsg_cy); }
	
	Rectangle flowGrid(Rectangle r, float iX, float iY, float countX, float countY) { return flowSplitV(flowSplitH(r, iX, countX), iY, countY); }
	Rectangle flowCenter(Rectangle r, float w, float h) { return Rectangle(r.origin.x+(r.size.width-w)/2, r.origin.y+(r.size.height-h)/2, w, h); }
	Rectangle flowCenter(Rectangle r, float s) { return Rectangle(r.origin.x+(r.size.width-s)/2, r.origin.y+(r.size.height-s)/2, s, s); }
	Rectangle flowCenterW(Rectangle r, float w) { return Rectangle(r.origin.x+(r.size.width-w)/2, r.origin.y, w, r.size.height); }
	Rectangle flowCenterH(Rectangle r, float h) { return Rectangle(r.origin.x, r.origin.y+(r.size.height-h)/2, r.size.width, h); }
	Rectangle flowTop(Rectangle r, float h) { return Rectangle(r.origin.x, r.origin.y, r.size.width, h); }
	Rectangle flowBottom(Rectangle r, float h) { return Rectangle(r.origin.x, r.origin.y+r.size.height-h, r.size.width, h); }
	Rectangle flowLeft(Rectangle r, float w) { return Rectangle(r.origin.x, r.origin.y, w, r.size.height); }
	Rectangle flowRight(Rectangle r, float w) { return Rectangle(r.origin.x+r.size.width-w, r.origin.y, w, r.size.height); }
	Rectangle flowWidth(Rectangle r, float w) { return Rectangle(r.origin.x, r.origin.y, w, r.size.height); }
	Rectangle flowHeight(Rectangle r, float h) { return Rectangle(r.origin.x, r.origin.y, r.size.width, h); }
	
	static RGBA darkenColor(RGBA color, float t) { return RGBA_new(cast(ubyte)(color.r*t),cast(ubyte)(color.g*t),cast(ubyte)(color.b*t),color.a); }
	static RGBA lightenColor(RGBA color, float t) { return RGBA_new(cast(ubyte)(color.r+(255-color.r)*t),cast(ubyte)(color.g+(255-color.g)*t),cast(ubyte)(color.b+(255-color.b)*t),color.a); }
	static RGBA blendColor(RGBA a, RGBA b, float t) { return RGBA_new(cast(ubyte)(a.r+(b.r-a.r)*t), cast(ubyte)(a.g+(b.g-a.g)*t), cast(ubyte)(a.b+(b.b-a.b)*t), cast(ubyte)(a.a+(b.a-a.a)*t)); }

	
	void styleKnobBasic(UIKnob knob) {
		//knob.colorBacking = correct(RGBA_new(0,0,0,0));
		//knob.margin = 0.05;
		knob.indicatorStyle = 0;
		knob.indicatorOrigin = 0;
		if (knob.indicatorStyle == 0) {
			knob.indicatorWidth = 0.1;
		} else {
			knob.indicatorWidth = 0.12;
			knob.indicatorLedCount = 25;
		}
		knob.indicatorGlow = 0.075;
		knob.indicatorGlowRadius=2;

		knob.knobRadius = 0.75;
		knob.hoverScale = 1.05;
		knob.colorKnob = __colorWhite;
		//knob.colorKnobFade = correct(RGBA_new(255,255,255,16));
		knob.colorKnobFade = correct(RGBA_new(10,22,55,32));
		knob.colorKnobFadeHover = lightenColor(knob.colorKnobFade, 0.5);//correct(RGBA_new(100,200,255,92))
		knob.knobFadeR1=0.33;
		knob.knobFadeR2=0.93;
		knob.knobFadeR3=0.98;
		knob.colorIndicator = __colorGrey4;
		knob.colorHover = __colorWhite;
		knob.colorDragged = __colorYellow;
		knob.labelColor = __colorWhite;
		knob.labelBackgroundColor = __colorClear; //__colorWhite.withAlpha(32);
		knob.labelValueColor = __colorGrey6;

		knob.jogStyle = 1;
		if (knob.jogStyle == 0) {
			knob.jogOffset = 0.75;
			knob.jogCircleRadius = 0.1;
		} else {
			knob.jogStyle = 1;
			knob.jogOffset = 0.9;
			knob.jogWidth = 0.125;
			knob.jogLength = 0.4;
		}
	}
	void styleKnobDark(UIKnob knob) {
		if(true) {
			knob.colorBacking = correct(RGBA_new(32,40,48,255));
			knob.indicatorStyle = 1;
			knob.colorIndicator = __colorGrey5;
			knob.colorIndicatorFill = __colorYellow;
			knob.labelColor = __colorWhite; //__colorBlack;
			knob.labelValueColor = __colorGrey2;
			knob.colorJog = __colorGrey2;
			//knob.indicatorWidth = 0.1;
			knob.margin = 0.1;
		} else {
			knob.colorBacking = correct(RGBA_new(0,16,32,64));
			knob.indicatorStyle = 0;
			knob.colorKnob = __colorGrey2;
			knob.colorKnobFade = correct(RGBA_new(0,0,0,32));
			knob.colorKnobFadeHover = darkenColor(knob.colorKnobFade, 0.5);//correct(RGBA_new(100,200,255,92));
			knob.colorIndicator = __colorGrey4;
			knob.colorIndicatorFill = correct(RGBA_new(255,0,255,255));
			knob.colorHover = __colorWhite;//.withAlpha(64);
			knob.colorHover.a = 64;
			knob.colorDragged = correct(RGBA_new(255,0,255,255));
			knob.labelColor = __colorBlack;
			knob.indicatorGlow = 0;
		}
	}
	
	void styleModule(string moduleName, RGBA color) {
	    color = __ModuleColors[getModuleParam(moduleName, "Color").intValue % 8];
	
		//getModulePanel(moduleName).color =  darkenColor(color, 0.1);
		
		getModulePanel(moduleName).color = __colorGrey2; // blendColor(darkenColor(color, 0.7), __colorGrey1, 0.8);
		
		auto tab = getModuleTab(moduleName);
		if (tab) {
			auto dColor = darkenColor(color,0.85);
			tab.colorFillInactive = __colorClear;
			tab.colorFillActive = __colorWhite;
			tab.colorFillHover = __colorWhite;
			tab.colorOutlineInactive = __colorGrey6;
			tab.colorOutlineActive = dColor;
			tab.colorOutlineHover = dColor;
			tab.colorLabelActive = dColor;
			tab.colorLabelHover = dColor;
			tab.styleOutlineThickness = 0.05;
			tab.labelOffsetY = -7*S;
			//tab.labelOffsetY = 0;
			//tab.labelRelativeHeight = 0.7;
			tab.labelRelativeHeight = 0;
		}
		
		auto indicator=getModuleIndicator(moduleName);
		if (indicator) {
			indicator.oneShot = true;
			indicator.peakHoldMS = 16.67*3;
			indicator.peakFallMS = 225;
			indicator.color = __colorGrey3; //darkenColor(color, 0.5);
			indicator.colorActive = darkenColor(color, 0.7);
			indicator.colorPeaking = lightenColor(color, 0.5);
			//if ( ["module3", "module4", "module5", "module6", "module10"].indexOf(moduleName)>=0 ) {
			if (!getModuleValue(moduleName, "Enabled").boolValue) {
				indicator.color = __colorGrey7;
				indicator.colorActive = __colorGrey7;
				indicator.colorPeaking = __colorGrey7;
			}
			indicator.scale = 0.8;
			indicator.scalePeaking = 0.97;
		}
		
		foreach (knobName; __ModuleKnobNames) {
			auto knob = getModuleKnob(moduleName, knobName);
			styleKnobBasic(knob);
			knob.colorIndicatorFill = color;
			knob.labelColor = color;
			knob.colorJog = darkenColor(color, 0.33);
		}
		
		foreach (knobName; ["Pan", "ToneNoise", "Bend", "FilterEnvBend"]) {
			auto knob = getModuleKnob(moduleName, knobName);
			knob.indicatorOrigin = 1;
			knob.indicatorDetent = true; //(knobName=="Bend");
		}
		
		foreach (knobName; ["GainDynamics", "BendDynamics", "FilterEnvBendDynamics"]) {
			auto knob = getModuleKnob(moduleName, knobName);
			knob.indicatorStyle = 1;
			knob.colorIndicatorFill = lightenColor(color, 0.75);
		}
		
		/*
		for (knobName; ["SendReverb"]) {
			auto knob = getModuleKnob(moduleName, knobName);
			knob.colorIndicatorFill = __colorPurple;
		}
		*/
		
		foreach (i; 1..4+1) {
			getModuleSeparator(moduleName, __Numbers[i]).color = __colorGrey9;
		}
		foreach (i; 1..5+1) {
			auto label = getModuleLabel(moduleName, __Numbers[i]);
			label.color = __colorGrey9;
			//label.colorShadow = correct(RGBA_new(0,0,0,96));
			label.shadowOffset = 0.1;
		}
		foreach (i; 1..3+1) {
			auto connector = getModuleConnector(moduleName, __Numbers[i]);
			connector.lineWidth = 3*S;
			connector.style = 0;
			connector.color = __colorGrey6;
			connector.colorEditing = __colorYellow;
		}
		/*
		foreach (i; 4..5+1) {
			auto connector = getModuleConnector(moduleName, __Numbers[i]);
			connector.style = 3;
		}
		*/
		{
			auto knob = getModuleKnob(moduleName, "Gain");
			knob.indicatorClipStart = __gainStartClip;
			knob.colorIndicatorClip = __colorClip;
		}
		
		foreach (buttonName; __ModuleButtonNames) {
			auto button = getModuleButton(moduleName, buttonName);
			button.color = __colorGrey5;
			button.colorBackgroundHover = __colorClear;
			button.colorHover = __colorPink;
			button.colorPressed = __colorYellow;
		}

			

	}
	
	void reflowModule(int index, float x, float y, float w, float h) {
		string moduleName = __ModuleNames[index];
		auto panel = getModulePanel(moduleName);
		if (panel) {
			panel.position = Rectangle(x,y,w,h);
		}
		
		auto indicator=getModuleIndicator(moduleName);
		if (indicator) {
			//auto tab = getModuleTab(moduleName)
			//if (tab) { indicator.position = flowCenter(tab.position, 50*S) }
			auto spacing = (w/__ModuleNames.length);
			auto indicatorX = (w/__ModuleNames.length)*index;
			auto indicatorSize = 50*S;
			auto offset = -indicatorSize/2;
			auto offsetX = spacing / 2;
			indicator.position = Rectangle(indicatorX + offsetX + offset, y - (32*S) + offset, indicatorSize, indicatorSize);
		}
		
		{
			auto fullR = flowPaddingH(Rectangle(x,y,w,h), 20*S);
			auto r = flowPaddingV(fullR, 40*S);
			
			auto gx = 11;
			auto gy = 2;
			auto gs = __knobSize*S;
			flowSetGrid(r, gx, gy);
			
			//get("module1ButtonIcon").position = Rectangle(0,0,10,10);
			getModuleButton(moduleName, "Icon").position = Rectangle(fullR.origin.x+10*S, fullR.origin.y, 30*S, 30*S);
			getModuleButton(moduleName, "Color").position = Rectangle(fullR.origin.x-20*S, fullR.origin.y, 30*S, 30*S);
			
			float ix = 0;
			getModuleKnob(moduleName, "Tone").position = flowCenter(flowUseGrid(ix, 0), gs);
			getModuleKnob(moduleName, "Bend").position = flowCenter(flowUseGrid(ix+1, 0), gs);
			getModuleKnob(moduleName, "OscWaveform").position = flowCenter(flowUseGrid(ix, 1), gs);
			getModuleKnob(moduleName, "BendDynamics").position = flowCenter(flowUseGrid(ix+1, 1), gs);
			
			ix = ix + 2.5;
			getModuleKnob(moduleName, "Noise").position = flowCenter(flowUseGrid(ix, 0), gs);
			getModuleKnob(moduleName, "FilterEnvBend").position = flowCenter(flowUseGrid(ix+1, 0), gs);
			getModuleKnob(moduleName, "LowpassQ").position = flowCenter(flowUseGrid(ix, 1), gs);
			getModuleKnob(moduleName, "FilterEnvBendDynamics").position = flowCenter(flowUseGrid(ix+1, 1), gs);
			
			ix = ix + 2.5;
			getModuleKnob(moduleName, "Decay").position = flowCenter(flowUseGrid(ix, 0), gs);
			getModuleKnob(moduleName, "Punch").position = flowCenter(flowUseGrid(ix, 1), gs);

			ix = ix + 1.5;
			getModuleKnob(moduleName, "Pan").position = flowCenter(flowUseGrid(ix, 0), gs);
			getModuleKnob(moduleName, "SendReverb").position = flowCenter(flowUseGrid(ix+1, 0), gs);
			getModuleKnob(moduleName, "ToneNoise").position = flowCenter(flowUseGrid(ix, 1), gs);
			getModuleKnob(moduleName, "Click").position = flowCenter(flowUseGrid(ix+1, 1), gs);

			ix = ix + 2.5;
			getModuleKnob(moduleName, "Gain").position = flowCenter(flowUseGrid(ix+1, 0), gs);
			getModuleKnob(moduleName, "Saturation").position = flowCenter(flowUseGrid(ix, 0.5), gs);
			getModuleKnob(moduleName, "GainDynamics").position = flowCenter(flowUseGrid(ix+1, 1), gs);

			auto sw = 2*S;
			auto sh = 300*S;
			getModuleSeparator(moduleName, "1").position = flowCenter(flowSplitH(r, 1.75, gx), sw, sh);
			getModuleSeparator(moduleName, "2").position = flowCenter(flowSplitH(r, 4.25, gx), sw, sh);
			getModuleSeparator(moduleName, "3").position = flowCenter(flowSplitH(r, 5.75, gx), sw, sh);
			getModuleSeparator(moduleName, "4").position = flowCenter(flowSplitH(r, 8.25, gx), sw, sh);
			auto cw = 10*S;
			auto ch = 130*S;
			getModuleConnector(moduleName, "1").position = flowCenter(flowSplitH(r, 1.6, gx), cw, ch);
			getModuleConnector(moduleName, "2").position = flowCenter(flowSplitH(r, 4.1, gx), cw, ch);
			getModuleConnector(moduleName, "3").position = flowCenter(flowSplitH(r, 10.6, gx), cw, ch);
			// getModuleConnector(moduleName, "4").position = flowTop(flowUseGrid(0.5, 0.05), 8*S);
			// getModuleConnector(moduleName, "5").position = flowTop(flowUseGrid(3.0, 0.05), 8*S);

			
			auto labelR = flowTop(flowPaddingV(fullR, 20*S), 25*S);
			getModuleLabel(moduleName, "1").position = flowCenterW(flowSplitH(labelR, 0.5, gx), 100*S);
			getModuleLabel(moduleName, "2").position = flowCenterW(flowSplitH(labelR, 3, gx), 100*S);
			getModuleLabel(moduleName, "3").position = flowCenterW(flowSplitH(labelR, 5, gx), 100*S);
			getModuleLabel(moduleName, "4").position = flowCenterW(flowSplitH(labelR, 7, gx), 100*S);
			getModuleLabel(moduleName, "5").position = flowCenterW(flowSplitH(labelR, 9.5, gx), 100*S);
		}
	}

	void parameterChanged() {
	    styleModules();
	}
	
	void styleModules() {
	    styleModule("module1", __colorRed);
	    styleModule("module2", __colorOrange);
	    styleModule("module3", __colorOrange);
	    styleModule("module4", __colorOrange);
	    styleModule("module5", __colorBlue);
	    styleModule("module6", __colorAqua);
	    styleModule("module7", __colorGreen);
	    styleModule("module8", __colorPink);
	    styleModule("module9", __colorPurple);
	    styleModule("module10", __colorYellow);
	}
	
	void styleUI() {
        //System.print("plugin.createUI()");
		//initLayout(); // Call from GUI.d
		
		(cast(UIColorPanel)get("background")).color = darkenColor(__colorAqua, 0.9); //__colorGrey7;
		//get("background").color = darkenColor(__ModuleColors[getModuleParam("module1", "Color").intValue % 8], 0.9);
		
        
		auto logo = cast(UILabel)get("logoLabel");
		logo.color = __colorGrey2;
        logo.colorShadow = __colorClear;
		logo.shadowOffset = 0.035;
		foreach (i; [ get("logoIcon"), get("simianIcon")]) {
			auto icon = cast(UIIcon)i;
			logo.color = __colorGrey2;
	        logo.colorShadow = __colorClear;
			logo.shadowOffset = 0.035;
		}
		(cast(UIIcon)get("simianIcon")).color = __colorGrey8;

        auto versionLabel = cast(UILabel)get("versionLabel");
        if (versionLabel) {
        	versionLabel.color = __colorGrey5;
        	//versionLabel.colorBackground = __colorWhite;
        }

		
		auto tabContainer = cast(UITabContainer)get("tabContainer");
		tabContainer.colorFill = __colorGrey9;
		
		foreach (moduleName; __ModuleNames) {
			getModulePanel(moduleName).color = __colorGrey5;
			}

		styleModules();
	    
	    //UI.root.color = correct(RGBA_new(255,255,255,255))

		foreach (b; [get("creditsButton"), get("upgradeButton")]) {
			auto button = cast(UIButton)b;
			button.colorBackground = __colorClear;
			button.colorBackgroundHover = __colorClear;
			button.colorBackgroundPressed = __colorClear;
			button.color = __colorWhite; //__colorGrey5; //.withAlpha(128); //Alpha not working on fonts/icons yet
			button.colorHover = __colorYellow;
			button.colorPressed = __colorYellow;
			button.colorShadow = blendColor(__colorAqua, __colorBlack, 0.333);
			button.shadowOffset = 0.075;
		}
		(cast(UIButton)get("upgradeButton")).color = __colorYellow;
		(cast(UIButton)get("upgradeButton")).colorHover = __colorAqua;


        //(cast(UIColorPanel)get("globalPanel")).color = __colorClear;
        (cast(UIColorPanel)get("globalPanel")).color = darkenColor(__colorRed, 0.8);
        //(cast(UIColorPanel)get("globalPanel")).color = darkenColor(__ModuleColors[getModuleParam("module1", "Color").intValue % 8], 0.8);

        
        foreach (knobName; __GlobalKnobNames) {
        	auto knob = getModuleKnob("global", knobName);
			styleKnobBasic(knob);
			styleKnobDark(knob);
        }
        auto knob = getModuleKnob("global", "Gain");
		knob.indicatorStyle = 0;
		knob.indicatorClipStart = __gainStartClip;
		knob.colorIndicatorClip = __colorClip;

        knob = getModuleKnob("global", "Drive");
		knob.indicatorStyle = 0;
		knob.indicatorClipStart = __gainStartClip;
		knob.colorIndicatorClip = __colorClip;

        knob = getModuleKnob("global", "Sensitivity");
		knob.colorIndicatorFill = __colorOrange;
		knob.indicatorStyle = 1;
		
        knob = getModuleKnob("global", "Saturation");
		knob.colorIndicatorFill = __colorPink;
		knob.indicatorStyle = 0;

        knob = getModuleKnob("global", "ReverbDecay");
		knob.colorIndicatorFill = __colorPurple;
		knob.indicatorStyle = 0;
		
        knob = getModuleKnob("global", "ReverbGate");
		knob.colorIndicatorFill = __colorPurple;
		knob.indicatorStyle = 0;

	    
	    auto resizer = cast(UIWindowResizer)get("resizerHint");
	    resizer.color = __colorGrey9;
	    resizer.colorHovered = __colorYellow;
	    resizer.colorDragged = __colorYellow;
	    resizer.colorCannotResize = __colorRed;
	    
	}


    void createUI() {
		styleUI();
    }

    void reflow() {
	    if(!initLayout())
	    	return;
    	styleUI();

		//System.print("plugin.reflow()");
		
      	
        get("background").position = Rectangle(0,0,W,H);
        get("logoIcon").position = Rectangle(10,10,60,60).scaleByFactor(S);
        get("logoLabel").position = Rectangle(70,15,160,50).scaleByFactor(S);
        get("simianIcon").position = Rectangle(802,467,180,180).scaleByFactor(S);
        auto versionLabel = get("versionLabel");
        if (versionLabel) {
        	versionLabel.position = Rectangle(70,58,160,22).scaleByFactor(S);
        }
		auto buttonSize = 35*S;
		get("creditsButton").position = Rectangle(240*S, 10*S, buttonSize, buttonSize);
        get("presetSelector").position = Rectangle(600,25,380,30).scaleByFactor(S);
        
        get("upgradeButton").position = Rectangle(0,515,112,140).scaleByFactor(S);

        //get("GG").position = Rectangle(DW-70, 6, 64, 64).scaleByFactor(S);
        //UI.getElementById("GG").position = Rectangle(DW-70, 6, 64, 64).scaleByFactor(S);
        //UIElement.new("GG").position = Rectangle(DW-70, 6, 64, 64).scaleByFactor(S);
        
	    (cast(UICreditsPanel)get("creditsPanel")).textSize = 25 * S;

        
        auto y = 80*S;
        auto tabHeight = 64*S;
        auto tabContainer = cast(UITabContainer)get("tabContainer");
		tabContainer.horizontalSkew = 0; //-20*S;
        tabContainer.position = Rectangle(0, y, W, tabHeight);

		y = y + tabHeight;
		auto panelHeight = 330*S;
		foreach (int i; 0..cast(int)__ModuleNames.length) {
			reflowModule(i, 0, y, W, panelHeight);
		}
		y = y + panelHeight;
		auto gs = __knobSize*S;
		
		panelHeight = H-y;
		auto fullR = Rectangle(0, y, W, panelHeight);
		auto r = fullR;
		get("globalPanel").position = r;
		//r = Rectangle(r.origin.x, r.origin.y-15*S, r.size.width, r.size.height);
		auto gx = 11;
		r = flowPaddingH(r, 20*S);
		flowSetGrid(r, gx, 1);
		auto offset = 1; //1.75;
        getModuleKnob("global", "Sensitivity").position = flowCenter(flowUseGrid(0.0+offset, 0), gs);
        getModuleKnob("global", "ReverbDecay").position = flowCenter(flowUseGrid(1.5+offset, 0), gs);
        getModuleKnob("global", "ReverbGate").position = flowCenter(flowUseGrid(2.5+offset, 0), gs);
        getModuleKnob("global", "Drive").position = flowCenter(flowUseGrid(4.0+offset, 0), gs);
        getModuleKnob("global", "Saturation").position = flowCenter(flowUseGrid(5.0+offset, 0), gs);
        getModuleKnob("global", "Gain").position = flowCenter(flowUseGrid(6.0+offset, 0), gs);
		auto sw = 2*S;
		auto sh = 140*S;
		getModuleSeparator("global", "1").position = flowCenter(flowSplitH(r, 0.75+offset, gx), sw, sh);
		getModuleSeparator("global", "2").position = flowCenter(flowSplitH(r, 3.25+offset, gx), sw, sh);
		auto labelR = flowTop(flowPaddingV(r, 5*S), 25*S);
		getModuleLabel("global", "1").position = flowCenterW(flowSplitH(labelR, 0.0+offset, gx), 100*S);
		getModuleLabel("global", "2").position = flowCenterW(flowSplitH(labelR, 2.0+offset, gx), 100*S);
		getModuleLabel("global", "3").position = flowCenterW(flowSplitH(labelR, 5.0+offset, gx), 100*S);

        get("resizerHint").position = Rectangle(W-30, H-30, 30, 30);
    }
}
