/*
   ┏━━━┓╱╱╱╱╱┏┓╱╱┏┓╱╱╱╱╱┏┓╱╱   
   ┃┏━┓┃╱╱╱╱╱┃┃╱╱┃┃╱╱╱╱╱┃┃╱╱╱  
   ┃┗━┛┣┓┏┳━━┫┃┏┓┃┃╱╱┏━━┫┗━┳━━┓
   ┃┏━━┫┃┃┃┏┓┃┗┛┃┃┃╱┏┫┏┓┃┏┓┃━━┫
   ┃┃╱╱┃┗┛┃┃┃┃┏┓┃┃┗━┛┃┏┓┃┗┛┣━━┃
   ┗┛╱╱┗━━┻┛┗┻┛┗┛┗━━━┻┛┗┻━━┻━━┛
    ━━━━━━━━━━━━━━━━━━━━━━━━━━ 

	Copyright (c) 2022 Punk Labs LLC

	This file is part of OneTrick SIMIAN

	OneTrick SIMIAN is free software: you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by the Free
	Software Foundation, either version 3 of the License, or (at your option)
	any later version.

	OneTrick SIMIAN is distributed in the hope that it will be useful, but WITHOUT
	ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
	FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
	more details.

	You should have received a copy of the GNU General Public License along with
	OneTrick SIMIAN.  If not, see <http://www.gnu.org/licenses/>.
*/

module punklabs.ui.helper;

import gui;
import core.atomic;
import dplug.core;
import dplug.gui;
import dplug.canvas;
import dplug.client;
import dplug.graphics.font;

import dplug.graphics.draw;
import std.math; //PI

import core.stdc.stdlib: malloc, realloc, free;
import core.stdc.stdio;
import core.stdc.string;

import punklabs.string.helper;



enum Icon
{
	//PunkIcons:
	PunkCircle,
	PunkKickDrum,
	PunkSnareDrum,
	PunkClaves,
	PunkDrumSticks,
	PunkTomDrumSmall,
	PunkTomDrumMedium, PunkTomDrum=PunkTomDrumMedium,
	PunkTomDrumLarge,
	PunkCymbal,
	PunkHihat,
	PunkOneTrickIcon,
	PunkSimianIcon,
	PunkUpgrade,
	//FontAwesome:
	None, PunkMAX=None,
	New,
	New2,
	Duplicate,
	Open,
	Save,
	Edit,
	Delete,
	Delete2,
	Refresh,
	Add,
	Remove,
	Accept,
	Cancel,
	Help,
	Info,
	Settings,
	Settings2,
	Settings3,
	Settings4,
	Settings5,
	ZoomIn,
	ZoomOut,
	Up,
	Down,
	Left,
	Right,
	Up2,
	Down2,
	Left2,
	Right2,
	Up3,
	Down3,
	Left3,
	Right3,
	Up4,
	Down4,
	Left4,
	Right4,
	Up5,
	Down5,
	Left5,
	Right5,
	Up6,
	Down6,
	Left6,
	Right6,
	KickDrum,
	SnareDrum,
	SnareRim,
	Wand,
	MagicWand,
	SteelDrum,
	GenericDrum,
	Rim,
	Cymbal,
	Hand,
	Clap,
	Clap2,
	Cow,
	Bell,
	Guitar,
	Disc,
	Favorite,
	Flag,
	Circle,
	Copyright,
	Trademark,
	Registered,
	Home,
	Cut,
	Share,
	Freeze,
	SquareWave,
	Volume1,
	Volume3,
	VolumeMute,
	Mic,
	MicMute,
	PaintBrush,
	PaintBrush2,
	PaintRoller,
	Spraypaint,
	Stamp,
	Palette,
	Background,
	Image,
	Fire,
	Screwdriver,
	Icons,
	MAX,
}

const string[Icon.MAX] IconChar = [
	//PunkIcons:
	"A",
	"B",
	"C",
	"D",
	"E",
	"F",
	"G",
	"H",
	"I",
	"J",
	"K",
	"L",
	"M",
	//FontAwesome:
	"",
	"📄",//📄🗋
	"",
	"",
	"📂",//📂🗁
	"",//💾
	"📝",
	"",//
	"⌫",
	"🗘",//🗘
	"",
	"",
	"",//
	"",//
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"🥁",//🥁
	"🔘",
	"",
	"",
	"",
	"",
	"",
	"",
	"✋",
	"",
	"",
	"",
	"🔔",
	"",
	"💿",
	"♥",//♥❤🤎💙💚💛🧡💜🖤🤍
	"",
	"",//🟠🟡🟢🟣🟤
	"",
	"",
	"",
	"",
	"✀",//✀✂✄
	"",
	"",
	"",
	"🔈",
	"🔊 ",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"🔥",
	"🪛",
	"",
];

/// Helper methods for drawing...
class UIHelper
{
public:
nothrow:
@nogc:

	static void initialize() // Ideally should be called early to avoid race conditions on first draw
	{
		ensureResources();
	}
	private static void ensureResources()
	{
		if (fontIcons is null)
			//fontIcons = mallocNew!Font(cast(ubyte[])(import("Fonts/icon-works_v1/icon-works-webfont.ttf")));
			fontIcons = mallocNew!Font(cast(ubyte[])(import("Fonts/Font-Awesome-6.x/webfonts/fa-solid-900.ttf")));
		if (fontPunkIcons is null)
			fontPunkIcons = mallocNew!Font(cast(ubyte[])(import("Fonts/PunkIcons/PunkIcons.ttf")));
	}
	
	static void fillColorRaw(ImageRef!RGBA rawMap, box2i[] dirtyRects, RGBA color)
	{
		if (color.a >= 255) //Overwrite, faster
		{
			foreach(dirtyRect; dirtyRects)
		    {
		        for (int y = dirtyRect.min.y; y < dirtyRect.max.y; ++y)
		        {
		            RGBA[] rawScan = rawMap.scanline(y);
		            rawScan[dirtyRect.min.x .. dirtyRect.max.x] = color;
		            /*
		            for (int x = dirtyRect.min.x; x < dirtyRect.max.x; ++x)
		            {
		                rawScan.ptr[x] = color;
		            }
		            */
		        }
		    }
		}
		else if(color.a > 0) //Blend, slower
		{
			foreach(dirtyRect; dirtyRects)
		    {
		        for (int y = dirtyRect.min.y; y < dirtyRect.max.y; ++y)
		        {
		            RGBA[] rawScan = rawMap.scanline(y);
		            for (int x = dirtyRect.min.x; x < dirtyRect.max.x; ++x)
		            {
		                rawScan.ptr[x] = blendColor(color, rawScan.ptr[x], color.a);
		            }
		        }
		    }
		}
	}

	static void drawLabelRaw(ImageRef!RGBA rawMap, box2i[] dirtyRects, box2i position, const(char)[] text, Font font, RGBA color, bool stableCenterY = true)
	{
		drawLabelRaw(rawMap, dirtyRects, cast(box2f)position, text, font, color, stableCenterY);
	}
	static void drawLabelRaw(ImageRef!RGBA rawMap, box2i[] dirtyRects, box2f position, const(char)[] text, Font font, RGBA color, bool stableCenterY = true)
	{
		if (color.a <= 0)
			return;
		float W = position.width;
		float H = position.height;
		vec2f textPosition = position.min+vec2f(W * 0.5f, H * 0.5f);
		float letterSpacingPx = 0;
		
		//float fontHeight = cast(float)font.measureText(fontTestFullString, H, 0).height;
		float fontAscent = font.getAscent(H);
		float fontDescent = font.getDescent(H);
		float fontHeight = fontAscent;//+fontDescent;
		
		float fontBaselineRatio = fontDescent/fontHeight;
		float fontSizeCorrection = H / fontHeight;
		float fontSizePx = H * fontSizeCorrection;
		
		if (stableCenterY)
			textPosition.y += H/2 / fontSizeCorrection;

		foreach(dirtyRect; dirtyRects)
		{
			auto croppedRaw = rawMap.cropImageRef(dirtyRect);
			
			/* //Debug Boundaries
			Canvas canvas;
			canvas.initialize(croppedRaw);
			canvas.translate(position.min.x-dirtyRect.min.x, position.min.y-dirtyRect.min.y); // Offset so we're using a consistent coordinate system
			canvas.fillStyle = RGBA(255,255,255,16);
			canvas.beginPath();
			canvas.moveTo(0,0);
			canvas.lineTo(W,0);
			canvas.lineTo(W,H);
			canvas.lineTo(0,H);
			//canvas.lineTo(0,0);
			canvas.closePath();
			canvas.fill();
			//*/
			
		    vec2f positionInDirty = textPosition - dirtyRect.min;
		    auto alignmentY = stableCenterY ? VerticalAlignment.baseline : VerticalAlignment.center;
			croppedRaw.fillText(font, text, fontSizePx, 0, color, positionInDirty.x, positionInDirty.y, HorizontalAlignment.center, alignmentY);
		}
	}

	static int countLines(const(char)[] text)
	{
		int lines = 1;
		for(int i=0; i < cast(int)text.length; i++)
		{
			if (text[i] == '\n')
				lines++;
		}
		return lines;
	}

	static void drawTextRaw(ImageRef!RGBA rawMap, box2i[] dirtyRects, box2i position, const(char)[] text, Font font, float fontSizePx, RGBA color, bool stableCenterY = true)
	{
		drawTextRaw(rawMap, dirtyRects, cast(box2f)position, text, font,  fontSizePx, color, stableCenterY);
	}
	static void drawTextRaw(ImageRef!RGBA rawMap, box2i[] dirtyRects, box2f position, const(char)[] text, Font font, float fontSizePx, RGBA color, bool stableCenterY = true)
	{
		//ulong init_size = strlen(text.ptr);
		char[] delim = cast(char[])"\n";
		char[] textcopy = StringHelper.duplicateString(text);
		char *ptr = strtok(textcopy.ptr, delim.ptr);
		box2f rect = position;
		rect.max.y = rect.min.y + fontSizePx;
		
		// Vertically center...
		int lineCount = countLines(text);
		float totalSize = fontSizePx*lineCount;
		float extraSize = position.height - totalSize;
		rect.min.y += extraSize/2;
		rect.max.y += extraSize/2;
		
		while(ptr != null)
		{
			//printf("'%s'\n", ptr);
			drawLabelRaw(rawMap, dirtyRects, rect, ptr[0..strlen(ptr)], font, color, stableCenterY);
			rect.min.y += fontSizePx;
			rect.max.y += fontSizePx;
			ptr = strtok(null, delim.ptr);
		}
		free(textcopy.ptr);
		
		/*
		int lineCount = text.length;
		box2i linePosition = position;
		linePosition.height = fontSizePx;
		for(int i=0; i<text.length; i++)
		{
			drawLabelRaw(rawMap, dirtyRects, linePosition, text[i], font, fontSizePx, color, stableCenterY);
			position.min.y += position.height;
		}
		*/
	}
	
	static void drawIconRaw(ImageRef!RGBA rawMap, box2i[] dirtyRects, box2i position, Icon icon, RGBA color, float scale = 1.0f)
	{
		drawIconRaw(rawMap, dirtyRects, cast(box2f)position, icon, color, scale);
	}
	static void drawIconRaw(ImageRef!RGBA rawMap, box2i[] dirtyRects, box2f position, Icon icon, RGBA color, float scale = 1.0f)
	{
		if (color.a <= 0)
			return;
		
		ensureResources();
			
		float W = position.width;
		float H = position.height;
		vec2f iconPosition = position.min+vec2f(W * 0.5f, H * 0.5f);
		float fontSizePx = H * scale;
		const(char)[] iconText = IconChar[icon];

		foreach(dirtyRect; dirtyRects)
		{
			auto croppedRaw = rawMap.cropImageRef(dirtyRect);
			
		    vec2f positionInDirty = iconPosition - dirtyRect.min;
			croppedRaw.fillText((icon<Icon.PunkMAX) ? fontPunkIcons : fontIcons, iconText, fontSizePx, 0, color, positionInDirty.x, positionInDirty.y, HorizontalAlignment.center, VerticalAlignment.center);
		}
	}
	
	// TODO: Save/Load the layout in some format to replace Wren in production
	static void saveLayout(UIElement rootElement, const(char)[] path)
	{
	}
	static void loadLayout(UIElement rootElement, const(char)[] path)
	{
	}
	
	private:
		static Font fontIcons = null;
		static Font fontPunkIcons = null;
		
		//static const string fontTestQuickString = "AHXYQjpgqp01(){}[]";
		//static const string fontTestFullString = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890(){}[]";

}
